# Simple Sample httpd/1.[01] server
# Stephen Uhler (c) 1996-1997 Sun Microsystems
# Httpd is a global array containing the global server state
#  root: the root of the document directory
#  port: The port this server is serving
#  listen:  the main listening socket id
#  accepts: a count of accepted connections so far
#  maxtime:     The max time (msec) allowed to complete an http request
#  maxused:     The max # of requests for a socket
# Modified: Tom Wilkason
#           Namespaced
#           support tml substitution
#           support vfs files

set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/HttpD.tcl,v 1.21 2002/12/31 04:58:12 wilkason Exp $
}
# HTTP/1.[01] error codes (the ones we use)
namespace eval HttpD {
   variable Httpd
   variable httpConfig
   variable HttpdUrlCache
   variable HttpMimeType
   variable HttpdErrors
   variable HttpdMimeType
   variable map
   variable localURL "/local"
   # Generic error response.
   variable HttpdErrorFormat {
      <title>Error: %1$s</title>
      Got the error: <b>%2$s</b><br>
      while trying to obtain <b>%3$s</b>
   }

   ;# Errors
   array set HttpdErrors {
      204 {No Content}
      400 {Bad Request}
      404 {Not Found}
      408 {Request Timeout}
      411 {Length Required}
      419 {Expectation Failed}
      503 {Service Unavailable}
      504 {Service Temporarily Unavailable}
   }

   ;# Default Settings
   array set Httpd {
      bufsize  50000
      maxtime  600000
      maxused  25
   }
   # convert the file suffix into a mime type
   # add your own types as needed
   array set HttpdMimeType {
      {}    text/plain
      .txt  text/plain
	   .tml	application/x-tcl-template
      .html text/html
      .htm  text/html
      .css  text/html
      .gif  image/gif
      .jpg  image/jpeg
      .png  image/png
   }
   ;# Initialize cgi map for encode/decode
   for {set i 1} {$i <= 256} {incr i} {
      set c [format %c $i]
      if {![string match \[a-zA-Z0-9\] $c]} {
         set map($c) %[format %.2X $i]
      }
   }
# These are handled specially
   array set map {
       " " +   \n %0d%0a
   }

}
# Start the server by listening for connections on the desired port.

proc HttpD::Server {root {port 80} {default index.html}} {
   Trace
   variable Httpd

   catch {close $Httpd(port)}  ;# it might already be running
   array set Httpd [list root $root default $default port $port]
   array set Httpd [list accepts 0 requests 0 errors 0]
   set Httpd(listen) [socket -server HttpD::Accept $port]
   return $Httpd(port)
}

# Accept a new connection from the server and set up a handler
# to read the request from the client. Called automatically
# On incoming request.

proc HttpD::Accept {sock ipaddr {port {}}} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data

   incr Httpd(accepts)
   Reset $sock $::snackAmpSettings(httpMaxused)
   Debug $sock Connect $ipaddr $port
}

# Initialize or reset the socket state

proc HttpD::Reset {sock left} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data

   array set data [list state start linemode 1 version 0 left $left]
   set data(cancel) [after $Httpd(maxtime) [list HttpD::Timeout $sock]]
   fconfigure $sock -blocking 0 -buffersize $Httpd(bufsize) \
         -translation {auto crlf}
   fileevent $sock readable [list HttpD::Read $sock]
}

# Read data from a client request. Called from a fileevent on the incoming
# socket.
# 1) read the request line
# 2) read the mime headers
# 3) read the additional data (if post && content-length not satisfied)

proc HttpD::Read {sock} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data
   # Use line mode to read the request and the mime headers
   if {$data(linemode)} {
      if {[catch {gets $sock line} readCount] } then {
         ;# Can get file busy on wrapped files, this
         ;# allows a retry.
         SocketDone $sock 1
         Debug $sock $readCount
         return
      }
      set state [string compare $readCount 0],$data(state)
      switch -glob -- $state {
         1,start {
            if {[regexp {(HEAD|POST|GET) ([^?]+)\??([^ ]*) HTTP/1.([01])} $line \
                     x data(proto) data(url) data(query) data(version)]} {
               set data(state) mime
               incr Httpd(requests)
               Debug $sock Request $data(left) $line
            } else {
               Error $sock 400 $line
            }
         }
         0,start {
            Debug $sock Warning "Initial blank line fetching request"
         }
         1,mime {
            if {[regexp {([^:]+):[  ]*(.*)}  $line {} key value]} {
               set key [string tolower $key]
               set data(key) $key
               if {[info exists data(mime,$key)]} {
                  append data(mime,$key) ", $value"
               } else {
                  set data(mime,$key) $value
               }
            } elseif {[regexp {^[   ]+(.+)} $line {} value] && \
                     [info exists data(key)]} {
               append data(mime,$data($key)) " " $value
            } else {
               Error $sock 400 $line
            }
         }
         0,mime {
            if {$data(proto) == "POST" && [info exists data(mime,content-length)]} {
               set data(linemode) 0
               set data(count) $data(mime,content-length)
               if {$data(version) && [info exists data(mime,expect]} {
                  if {$data(mime,expect) == "100-continue"} {
                     puts $sock "100 Continue HTTP/1.1\n"
                     flush $sock
                  } else {
                     Error $sock 419 $data(mime,expect)
                  }
               }
               fconfigure $sock -translation {binary crlf}
            } elseif {$data(proto) != "POST"}  {
               Respond $sock
            } else {
               Error $sock 411 "Confusing mime headers"
            }
         }
         -1,* {
            if {[eof $sock]} {
               Debug $sock Error "Broken connection fetching request, client may have disconnected"
               SocketDone $sock 1
            } else {
               Debug $sock "Partial read, retrying"
            }
         }
         default {
            Error $sock 404 "Invalid http state: $state,[eof $sock]"
         }
      }

      # Use counted mode to get the post data

   } elseif {![eof $sock]} {
      append data(postdata) [read $sock $data(count)]
      set data(count) [expr {$data(mime,content-length) - \
               [string length $data(postdata)]}]
      if {$data(count) == 0} {
         Debug $sock "Respond 0 data count"
         Respond $sock
      }
   } else {
      Log $sock Error "Broken connection reading POST data"
      SocketDone $sock 1
   }
}

# Done with the socket, either close it, or set up for next fetch
#  sock:  The socket I'm done with
#  close: If true, close the socket, otherwise set up for reuse

proc HttpD::SocketDone {sock close} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data
   after cancel $data(cancel)
   set left [incr data(left) -1]
   unset data
   if {$close} {
      catch {close $sock}
      Debug $sock "Closing socket $sock"
   } else {
      Reset $sock $left
      Debug $sock "Resetting socket $sock - $left bytes left"
   }
   return ""
}

# A timeout happened

proc HttpD::Timeout {sock} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data
   Error $sock 408
}

###############################################################################
# Function   : ContentHeader
# Description: Return content header for some file
# Author     : Tom Wilkason
# Date       : 4/28/2002
###############################################################################
proc HttpD::ContentHeader {sock mypath} {
   Trace
   upvar #0 Httpd$sock data
   append result "HTTP/1.$data(version) 200 Data follows\n"
   append result "Date: [Date [clock seconds]]\n"
   append result "Last-Modified: [Date [file mtime $mypath]]\n"
   append result "Content-Type: [ContentType $mypath]\n"
   append result "Content-Length: [file size $mypath]"
   return $result
}
# Handle file system queries.  This is a place holder for a more
# generic dispatch mechanism.
proc HttpD::Respond {sock} {
   Trace
   variable Httpd
   variable HttpdUrlCache
   variable localURL
   upvar #0 Httpd$sock data
   if {![info exists data(query)]} {
      set data(query) {}
   }
   ;# Get url with http stuff striped off
   regsub  -- {(^http://[^/]+)?} $data(url) {} url
   if {[info exists HttpdUrlCache($url)]} {
      set mypath $HttpdUrlCache($url)
      Debug $sock "Using cached $mypath"
   } else {
      ;# Translate any "/local" requests to a user specified location
      if {![regsub "^${localURL}(.*)" $url "$::snackAmpSettings(HttpLocal)\\1" mypath]} {
         ;# If no /local found, then translate file to complete path from true root
         set mypath [Url2File $Httpd(root) $url]
      }
      ;# If root specified, the assume as remote control
      if {$url=="/"} {
         set mypath "[string trimright $mypath /]/$Httpd(default)"
      } elseif {[file isdirectory $mypath]} {
         ;##
         ;# Attempt default file names
         ;#
         foreach {attempt} {
            default.html
            default.htm
            index.html
            index.htm
         } {
            set try "[string trimright $mypath /]/$attempt"
            if {[file exists $try]} {
               set mypath $try
               break
            }
         }
      }
      set HttpdUrlCache($url) $mypath
   }
   Debug $sock "url:'$data(url)' - '$url' - '$mypath'"
   ;##
   ;# Now check the path
   ;#
   if {[string length $mypath] == 0} {
      Error $sock 400
   } elseif {![file readable $mypath]} {
      ;##
      ;# Treat request as internal dispatch if the
      ;# url decoded into a local file did not exist.
      #/control/?GLA=K106&Date=2000-04&submit=Query

      regexp {/(.+[^/])} $data(url) xx name
      Debug "Query: '$data(query)'=''$mypath"
      ;# Grab the appropriate query parameters
      switch -- $data(proto) {
         GET {set param $data(query)}
         POST {set param $data(postdata)}
         default {set param ""}
      } ;# End Switch
      #puts "'$data(proto)' param '$param'"
      if {$::snackAmpSettings(EnableRemoteControl)} then {
         if {[catch {::${name}::process $sock $param} result] } then {
            ;# data will disappear here, socket is closed
            Error $sock 400 "Unable to process $result\n$::errorInfo"
         } else {
            Debug $sock "Processed: $mypath"
         }
      } else {
         puts $sock "<body bgcolor='bisque'>\n<html>\n\t<h1>Remote Control disabled for [info hostname]</h1>\n</html>\n</body>"
         HttpD::SocketDone $sock 1
      }
   } else {
      ;##
      ;# File exists, if it is a tml macro file, process it
      ;# and write out corresponding html file.
      ;# TFW: Improve on this such that we don't have to write the file
      ;# out first, will need new routine to build content header.
      if {$::snackAmpSettings(EnableRemoteControl) && [string equal -nocase [file extension $mypath] ".tml"]} {
         Debug $sock "Processing TML File $mypath"
         ;# Open tml
         if {[catch {set fid [open $mypath r]} error] } then {
            Log $sock "TML Open Error:$errorInfo"
         } else {
            #fconfigure $fid -translation binary
            set contents [read $fid]
            close $fid
            ;# Substitute and rename output file
            ;# TFW: If any $xxx stuff in file, the subst will fail.
            if {[catch {subst $contents} contents] } then {
               Log $sock "TML Error: $::errorInfo"
            } else {
               ;# translate to html file
               set mypath "[file join $::HOME [file root [file tail $mypath]].html]"
               ;# Write out to html file
               set fid [open $mypath w]
               fconfigure $fid -translation binary
               puts -nonewline $fid $contents
               close $fid
            }
         }
      }
      ;# Feed proper html header to browser
      puts $sock [ContentHeader $sock $mypath]
      ## Should also close socket if recvd connection close header
      set close [expr {$data(left) == 0}]
      if {$close} {
         puts $sock "Connection close:"
      } elseif {$data(version) == 0 && [info exists data(mime,connection)]} {
         if {$data(mime,connection) == "Keep-Alive"} {
            set close 0
            puts $sock "Connection: Keep-Alive"
         }
      }
      catch {
         puts $sock ""
         flush $sock
      }
      ;##
      ;# Process complete html files here copying from file
      ;# to the socket. 
      ;#
      if {$data(proto) != "HEAD"} {
         ;# TFW; Here is where we copy html to output
         if {[catch {
            set in [open $mypath r]
            fconfigure $sock -translation binary
            fconfigure $in -translation binary
            fcopy $in $sock -command [list HttpD::CopyDone $in $sock $close]
         } error] } then {
            Log $sock "Respond: $error"
         }
      } else {
         SocketDone $sock $close
      }
   }
}
;##
;# Load
;#
proc HttpD::LoadService { sock service } {
   Trace
   upvar #0 config$sock data
   variable httpConfig
   if {$service == ""} {
      return 0
   }
   if {![info exists httpConfig(services)] || [lsearch -exact $httpConfig(services) $service ] == -1 } {
      Debug "$sock Info Loading service: $service"
      if {[catch {namespace eval $service init} result] } then {
         Error $sock 503 "unable to start service \"$service\"\n$result"
         return 0
      } else {
         lappend httpConfig(services) $service
      }
   }
   return 1
}

# Callback when file is done being output to client
# in:  The fd for the file being copied
# sock: The client socket
# close: close the socket if true
# bytes: The # of bytes copied
# error:  The error message (if any)

proc HttpD::CopyDone {in sock close bytes {error {}}} {
   Trace
   variable Httpd
   upvar #0 Httpd$sock data
   close $in
   Debug $sock Done $bytes bytes
   SocketDone $sock $close
}
#
# Return the content type based on file name extension
#
proc HttpD::ContentType {path} {
   Trace
   variable HttpdMimeType

   set type text/plain
   catch {set type $HttpdMimeType([file extension $path])}
   return $type
}

# Respond with an error reply
# sock:  The socket handle to the client
# code:  The httpd error code
# args:  Additional information for error logging

proc HttpD::Error {sock code args} {
   Trace
   upvar #0 Httpd$sock data
   variable Httpd
   variable HttpdErrors
   variable HttpdErrorFormat
   if {![info exists data(version)]} {
      ;# This handles cases where buttons are clicked too fast, just bail out
      return
   } ;# end if
   append data(url) ""
   incr Httpd(errors)
   set message [format $HttpdErrorFormat $code $HttpdErrors($code) $data(url)]
   append head "HTTP/1.$data(version) $code $HttpdErrors($code)"  \n
   append head "Date: [Date [clock seconds]]"  \n
   append head "Connection: close"  \n
   append head "Content-Length: [string length $message]"  \n

   # Because there is an error condition, the socket may be "dead"

   catch {
      fconfigure $sock  -translation crlf
      puts -nonewline $sock $head\n$message
      flush $sock
   } reason
   SocketDone $sock 1
   Log $sock Error $code $HttpdErrors($code) $args $reason
}
# Log an Httpd transaction.
# This should be replaced as needed.

proc HttpD::Log {sock args} {
   #Trace
   saLog "LOG: $sock $args"
}
proc HttpD::Debug {sock args} {
   if {$::snackAmpSettings(Trace)} {
      saLog "Debug: $sock $args"
   }
}

# Generate a date string in HTTP format.

proc HttpD::Date {seconds} {
   return [clock format $seconds -format {%a, %d %b %Y %T %Z}]
}


# Convert a url into a pathname. (UNIX version only)
# This is probably not right, and belongs somewhere else.
# - Remove leading http://... if any
# - Collapse all /./ and /../ constructs
# - expand %xx sequences -> disallow "/"'s  and "."'s due to expansions

proc HttpD::Url2File {root url} {
   Trace
   regsub -all  -- {//+} $url / url     ;# collapse multiple /'s
   while {[regsub -all  -- {/\./} $url / url]} {}   ;# collapse /./
   while {[regsub -all  -- {/\.\.(/|$)} $url /\x81\\1 url]} {} ;# mark /../
   while {[regsub  -- "/\[^/\x81]+/\x81/" $url / url]} {} ;# collapse /../
   if {![regexp "\x81|%2\[eEfF]" $url]} { ;# invalid /../, / or . ?
      set result $root[CgiMap $url]
   } else {
      set result ""
   }
   return $result
}

# Decode url-encoded strings.

proc HttpD::CgiMap {data} {
   Trace
   regsub -all  -- {([][$\\])} $data {\\\1} data
   regsub -all  -- {%([0-9a-fA-F][0-9a-fA-F])} $data  {[format %c 0x\1]} data
   return [subst $data]
}
;##
;# Not used
;#
proc HttpD::GetParams { vars args } {
   Trace
   upvar args arglist

   if { [ llength $arglist ] == 1 } {
      # braced set of args
      eval set arglist $arglist
   }
   foreach { var val } $vars {
      uplevel 1 set $var \"$val\"   ;#"
   }
   foreach arg $arglist {
      set param [ split $arg "=" ]
      set var [ lindex $param 0 ]
      set val [ join [ lreplace $param 0 0 ] "=" ]
      if { [ lsearch $vars $var ] != -1 } {
         uplevel 1 set $var \{$val\}
      }
   }
}
# HttpD::Decode
#
#  This decodes data in www-url-encoded format.
#
# Arguments:
#  An encoded value
#
# Results:
#  The decoded value
proc HttpD::Decode {data} {
   Trace
    regsub -all {\+} $data " " data
    regsub -all {([][$\\])} $data {\\\1} data
    regsub -all {%([0-9a-fA-F][0-9a-fA-F])} $data  {[format %c 0x\1]} data
    return [subst $data]
}
# HttpD::Encode
#
#  This encodes data in www-url-encoded format.
#
# Arguments:
#  A string
#
# Results:
#  The encoded value

proc HttpD::Encode {string} {
   Trace
   variable map
   regsub -all \[^a-zA-Z0-9\] $string {$map(&)} string
   regsub -all \n $string {\\n} string
   regsub -all \t $string {\\t} string
   regsub -all {[][{})\\]\)} $string {\\&} string
   return [subst $string]
}
###############################################################################
# Function   : HttpD::Start
# Description: Start up a local webserver on the next free port starting at 8080
#               Save the port number                                             
# Author     : Tom Wilkason
# Date       : 4/14/2002
###############################################################################
proc HttpD::Start {port root default} {
   Trace
   global TA
   set tries 0
   while {[catch {HttpD::Server $root $port $default}] && $tries <= 20} {
      incr port
      incr tries
   }
   if {$tries == 20} {
      Log "Failed to start the server"
   } else {
      Log "Starting Tcl httpd server on [info hostname] port $port"
   }
   return $port
}
package provide HttpD 1.0
##################################################################
##################################################################
##################################################################
##################################################################
#####    E N D  O F   W E B   G E N E R I C   S E R V E R    #####
##################################################################
##################################################################
##################################################################
##################################################################
##################################################################
;##
;# justcontrol namespace to control the SnackAmp player from
;# a web browser. When user enters /justcontrol ... in to the browser
;# The process function of this namespace will be called passing any
;# user parameters. justcontrol does not send anything back to
;# the client. It is very quick processing.
;#
namespace eval justcontrol {
   proc init { } {
     return 1
   }
}
proc justcontrol::process {sock {params ""}} {
   Trace
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         bt_Level {
            ;# Would be nice not to have to push data back
            ;# maybe hold to end to push it out if needed.
            if {[catch {::macros::bumpLevel $Cmd} result] } then {
            }
         }
         default {}
      } ;# End Switch
   } ;# end foreach
   return 0
}

;##
;# control namespace to control the SnackAmp player from
;# a web browser. When user enters /control ... in to the browser
;# The process function of this namespace will be called passing any
;# user parameters. To add a new command then create another namespace
;# and provide an init and process function.
;#
namespace eval ::control {
   variable Details
   variable LastFile {xxxxx}
   variable TagIDs {
      Active        Active
      "File Name"   Name
      Folder        Folder
      Artist        Artist
      Title         Title
      Album         Album
      Track         Track
      Media         Media
      Year          Year
      Genre         Genre
      Note          Note
      Comp          Comp
      Encoding      Enc
      Desc          Desc
      Size          fSize
      Date          fDate
      Duration      duration
      Position      Position
      "Sample Rate" rate
      Gain          levelFactor
   }
     ;# <meta http-equiv="Content-Type" content="text/html; charset=windows-1252">
   variable head {
<html>
   <head>
      <meta http-equiv="Content-Language" content="en-us">
      <meta http-equiv="pragma"           content="no-cache">
      <meta name="GENERATOR"              content="SnackAmp Web Server">
      <meta name="ProgId"                 content="SnackAmp.Document">
      <LINK REL=StyleSheet                HREF="/snackamp.css"         TYPE="text/css" MEDIA=screen>
      <title>$title</title>
   }
   variable songTemplate {
      <table border="1" width="100%" height="50" cellpadding="0" cellspacing="0" bgcolor="#333300">
        <tr>
          <td width="9%" class="songTemplateLabel">Title</td>
          <td width="91%" colspan="5"><b><font size="4" color="lightblue">$Details(Title)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel" width="9%">Artist</td>
          <td class="songTemplateInfo"  width="91%" colspan="5">$Details(Artist)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel" class="songTemplateLabel" width="9%">Album</td>
          <td class="songTemplateInfo" width="91%" colspan="5">$Details(Album)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel" width="9%">File</td>
          <td class="songTemplateInfo" width="91%" colspan="5">$Details(Name)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel" width="9%">Folder</td>
          <td class="songTemplateInfo" width="91%" colspan="5">$Details(Folder)</td>
        </tr>
        <tr>
          <td width="9%"   class="songTemplateLabel">Size</td>
          <td width="30%"  class="songTemplateInfo">$Details(fSize)</td>
          <td width="13%"  class="songTemplateLabel">Genre</td>
          <td width="20%"  class="songTemplateInfo">$Details(Genre)</td>
          <td width="10%"  class="songTemplateLabel">Gain</font></i></b></td>
          <td width="100%" class="songTemplateInfo">[id3Tag::levelText $Details(levelFactor)]</td>
        </tr>
        <tr>
          <td width="9%"   class="songTemplateLabel">Date</td>
          <td width="30%"  class="songTemplateInfo">$Details(fDate)</td>
          <td width="13%"  class="songTemplateLabel">Rate</td>
          <td width="20%"  class="songTemplateInfo">$Details(rate)</td>
          <td width="10%"  class="songTemplateLabel">Track</td>
          <td width="100%" class="songTemplateInfo">$Details(Track)</td>
        </tr>
        <tr>
          <td width="9%"   class="songTemplateLabel">Year</td>
          <td width="30%"  class="songTemplateInfo">$Details(Year)</td>
          <td width="13%"  class="songTemplateLabel">Duration</td>
          <td width="20%"  class="songTemplateInfo">$Details(duration)</td>
          <td width="10%"  class="songTemplateLabel">Position</td>
          <td width="100%" class="songTemplateInfo">$Details(Position)</td>
        </tr>
      </table>
   }
   variable row_colors
   array set row_colors {
      0 {"white"}
      1 {"lightyellow"}
   }
   proc init { } {
     return 1
   }
}
###############################################################################
# Function   : control::process state machine
# Description: Main function that is called by web server when /control
#              is entered. Inside is a loop that iterates over each parameter
#              on the POST/GET and handles each one.
#              e.g. /control?a=2&b=c  <= two params a= and b=
# Author     : Tom Wilkason
# Date       : 4/24/2002
###############################################################################
proc control::process {sock {params ""}} {
   Trace
   #puts "control::process $sock=>$params"
   ;##
   ;# Replace puts with appending to string then doing as a batch if needed
   ;#
   puts $sock [control::leader "SnackAmp Control" "/control?Status=Refresh+Data"]
   puts $sock $jukebox::javascript
   #puts $sock "$params"
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         bt_Control {
            if {[catch {::macros::$Cmd} result] } then {
               puts $sock "<p>$::errorInfo</p>"
            }
            after 50 set ::control::flag 1
            vwait ::control::flag
         }
         bt_Level {
            if {[catch {::macros::bumpLevel $Cmd} result] } then {
               puts $sock "<p>$::errorInfo</p>"
            }
         }
         play {
            ;# Play a specific song
            ;# /control?play=some+name.mp3
            if {[catch {playSpecificSong $Cmd} error] } then {
               puts stderr $result
               puts $sock $result
            }
            after 50 set ::control::flag 1
            vwait ::control::flag
         }
         delete {
            LB::delete $Cmd
            updateFileDisplay
            ;# Remove from playlist
         }
         cmb_APL {
            ;# Select an auto-playlist file
            ;# TFW: If a real + is in file name, this may fail
            macros::readAPLFile [file join $::HOME $Cmd.tafs] 1
         }
         cmb_Sort {
            ;# Sort/Shuffle the files
            switch -glob -- [string tolower $Cmd] {
               sort*tracks     {set ::snackAmpSettings(sortMode) 0}
               shuffle*tracks  {set ::snackAmpSettings(sortMode) 1}
               shuffle*sort    {set ::snackAmpSettings(sortMode) 2}
               shuffle*shuffle {set ::snackAmpSettings(sortMode) 3}
            } ;# End Switch
            filtering::sortMusic
         }
         filter_text {
            ;# stage 1
            set ::TA(filter) $Cmd
            ;# Note: These will happen in this order, search_replace can process
            ;# Button=Filter&filter_text=ab&filter_what=All+Files&search_replace=Replace
         }
         filter_what {
            ;# stage 2
            switch -glob -- [string tolower $Cmd] {
               "all*files" {set Search_What 1}
               "play*" {set Search_What playList}
               "auto*" {set Search_What 0}
               default {set Search_What 1}
            } ;# End Switch
         }
         search_replace {
            ;# stage 3, do the filtering
            if {![info exists Search_What]} {set Search_What 1}
            set Search_Replace [string tolower $Cmd]
            filtering::pushFilter
            macros::filterMusic $Search_Replace $Search_What
            ;# stage 3
         }
         cmb_PL {
            ;# Stage 1
            ;# Set the playlist file
            set Playlist_File [file join $::HOME $Cmd.taf]
         }
         PL_search_replace {
            ;# Stage 2
            ;# And then apply it
            macros::ProcessPlaylistFile $Playlist_File [string tolower $Cmd]
         }
         default {}
      } ;# End Switch
   } ;# end foreach
   ;# Wait to let player take hold
   puts $sock [sndInfo]
   puts $sock [nextToPlay 1 $::snackAmpSettings(httpNextLength)]
   puts $sock "</body>\n</html>"
   HttpD::SocketDone $sock 1
   return 0
}
###############################################################################
# Function   : leader
# Description: Page leader for dymanic page
# Author     : Tom Wilkason
# Date       : 4/24/2002
###############################################################################
proc control::leader {title {url ""}} {
   Trace
   variable head
   ;# title is used in subst
   append result [subst $head]
   if {[string length $url]>0} {
     append result \n
     append result [subst {    <META HTTP-EQUIV="Refresh" CONTENT="$::snackAmpSettings(refreshRate);URL=$url">}]
     append result \n
   } ;# end if
   append result "  </head>\n"
   append result "<body>\n"
}
###############################################################################
# Function   : trackInfo
# Description: Return, in a table, the details on some file
#
# Author     : Tom Wilkason
###############################################################################
proc control::trackInfo {fileName} {
   variable TagIDs
   foreach {name tag} $TagIDs {
      set Details($tag) ""
   }
   set tempSound [snack::sound -debug 0 -file $fileName]
   if {[catch {id3Tag::songDetails $fileName $tempSound Details} error] } then {
      HttpD::Log "sndInfo:$error"
      $tempSound destroy
      return $error
   }
   $tempSound destroy
   set Details(Folder) [file dirname $fileName]
   set Details(Name) [file tail $fileName]
   set Details(Position) "N/A"
   ;# TFW:This forces table borders on blank cells, there may be a way
   ;# to do this in the table definition.
   foreach field [array names Details] {
      if {[string length $Details($field)]==0} {
         set Details($field) "&nbsp;"
      }
   } ;# end foreach
   set result [control::songTable Details]
}
###############################################################################
# Function   : sndInfo
# Description: Return, in a table, the details on the current song
#
# Author     : Tom Wilkason
###############################################################################
proc control::sndInfo {args} {
   Trace
   variable LastFile
   variable TagIDs
   variable Details
   set fileName [soundControl::fileName]
   ;##
   ;# If the file name hasn't changed, don't
   ;# bother requerying the static song data.
   ;#
   if {![string equal $fileName $LastFile]} {
      foreach {name tag} $TagIDs {
         set Details($tag) ""
      }
      if {[catch {id3Tag::songDetails $fileName s Details} error] } then {
         HttpD::Log "sndInfo:$error"
         return
      }
      set Details(Folder) [file dirname $fileName]
      set Details(Name) [file tail $fileName]
      set LastFile $fileName
   }
   set Active [soundControl::playState]

   set Details(Position) [clock format [expr {round([soundControl::position])}]  -format %M:%S]
   append result "<h2>SnackAmp currently $Active on [info hostname]:$::HttpD::Httpd(port)</h2>"
   ;# TFW:This forces table borders on blank cells, there may be a way
   ;# to do this in the table definition.
   foreach field [array names Details] {
      if {[string length $Details($field)]==0} {
         set Details($field) "&nbsp;"
      }
   } ;# end foreach
   append result [control::songTable Details]
   append result "<br>\n"
   return $result
}
###############################################################################
# Function   : nextToPlay
# Description: Print out the next N tracks in the playlist
#
# Author     : Tom Wilkason
###############################################################################
proc control::nextToPlay {playButton {N 20}} {
   Trace
   global LB
   variable row_colors
   set result {}
   if {[catch {
      set top [expr {$LB::indexPlaying-5}]
      set end [LB::index end]
      set total [expr {$end - [LB::active]}]
      set newEnd [expr {$top+$N}]
      if {$newEnd > $end} {set newEnd $end} ;# end if
      if {$top < 0} {set top 0} ;# end if
      #set fnt1 {<font face="Helvetica">}
      #set fnt2 {</font>}
      set current [file tail $::TA(currentMF)]
      append result "<h2>Next [expr {$newEnd-$top}] Queued Tracks (of [commify $end] Tracks in Playlist)</h2>"
      append result {<table bgcolor="ghostwhite"  border="1" width ="100%" cellpadding="0" cellspacing="0" >}
      for {set i $top} {$i < $newEnd} {incr i} {
         ;# <a href="/control?play=Artie+Shaw+-+Begin+The+Beguine.mp3">Artie Shaw - Begin The Beguine.mp3</a>
         set bgc $row_colors([expr {($i/5)&0x1}])
         set song [LB::get $i]
         set file [LB::getFullName $i 0]
         set track [file root [file tail $file]]
         set folder [file tail [file dirname $file]]
         if {$folder==""} {
            set folder "&nbsp;"
         } ;# end if
         set trackInfo "<td>$folder</td><td>$track</td>"
         set imRef [jukebox::questionButton $song 200 "info"]
         if {$playButton} {
            append imRef [subst {<td><a href="/control?play=[HttpD::Encode $song]"><img border="0" src="/remote/images/playM.gif"></a></td>}]
            append imRef [subst {<td><a href="/control?delete=$i"><img border="0" src="/remote/images/deleteM.gif"></a></td>}]
         }
         if {[string match "*$song*" $::TA(currentMF)]} {
            set bgc {"yellow"}
         }
         append result "<tr bgcolor=$bgc><td>$i</td>${imRef}${trackInfo}</tr>\n"
      } ;# end for
      append result </table>
   } error]} {
      return $error
   } else {
      return $result
   }
}

###############################################################################
# Function   : ::control::aplFiles
# Description: Callback from TML file to return a list of all auto-playlist files
# Author     : Tom Wilkason
# Date       : 4/28/2002
###############################################################################
proc control::aplFiles {args} {
   Trace
   set result ""
   if {![catch {set Files [lsort -dictionary [glob [file join $::HOME *.tafs]]]}]} {
      ;# Create option list of all auto-playlist files
      foreach {file} $Files {
         set choice [file root [file tail $file]]
         if {[string equal $choice $::TA(AutoPlayListSelection)]} {
            append result "<option selected>$choice</option>\n"
         } else {
            append result "<option>$choice</option>\n"
         }
      } ;# end foreach
   }
   return [string trim $result]
}
###############################################################################
# Function   : ::control::plFiles
# Description: Callback from TML file to return a list of all playlist files
# Author     : Tom Wilkason
# Date       : 4/28/2002
###############################################################################
proc control::plFiles {args} {
   Trace
   set result ""
   if {![catch {set Files [lsort -dictionary [glob [file join $::HOME *.taf]]]}]} {
      ;# Create option list of all auto-playlist files
      foreach {file} $Files {
         set choice [file root [file tail $file]]
         if {[string equal $choice $::TA(PlayListSelection)]} {
            append result "<option selected>$choice</option>\n"
         } else {
            append result "<option>$choice</option>\n"
         }
      } ;# end foreach
   }
   return [string trim $result]
}
###############################################################################
# Function   : ::control::songTable
# Description: Return HTML for the song details
# Author     : Tom Wilkason
# Date       : 4/29/2002
###############################################################################
proc control::songTable {_Details} {
   upvar $_Details Details
   variable songTemplate
   set result [subst $songTemplate]
   return $result
}


namespace eval display {
   proc init { } {
     return 1
   }
}
###############################################################################
# Function   : display
# Description: Main function that is called by web server when /display
#              is entered.
#              e.g. /display
# Author     : Tom Wilkason
# Date       : 12/15/2002
###############################################################################
proc display::process {sock {params ""}} {
   Trace
   puts $sock [control::leader "SnackAmp Playlist" "/display"]
   puts $sock $jukebox::javascript
   ;# Wait to let player take hold
   after 50 set ::control::flag 1
   vwait ::control::flag
   puts $sock $jukebox::close_window
   puts $sock [control::sndInfo]
   puts $sock [control::nextToPlay 0 $::snackAmpSettings(httpNextLength)]
   ##puts $sock "<br><a href=/jukebox>Request Tracks to Add To Playlist</a>"
   puts $sock "</body>\n</html>"
   HttpD::SocketDone $sock 1
   return 0
}
###############################################################################
# Function   : jukebox namespace
# Description: Handles interface to request songs to be added to playlist
#              
# Author     : Tom Wilkason
###############################################################################
namespace eval jukebox {
   variable search_form {
    <!-- Search Database -->
    <table border="1" width="100%" bordercolor="darkblue" cellspacing="0" cellpadding="0">
      <tr>
        <td width="100%" bgcolor="darkblue" align="center"><font color="white" face="Helvetica">
        <b>Search All Tracks for Matching Text</b></font></td>
      </tr>
      <tr>
        <td width="100%" align="left">
          <form method="POST" action="/jukebox" TITLE="Enter text then hit Search to find all songs that contain the text">
            <input type="submit" value="Search" name="Button">
            <font face="Helvetica">
              <input type="text" name="jb_search" size="30">
            </font>
          </form>
        </td>
      </tr>
    </table>
    <p>
   }
   variable close_window {<a href="" onClick="window.close()">Click this to close the window</A>}
   ;##
   ;# Javascript helper routines
   ;#
   variable javascript {
     <script language="JavaScript">
     //<!--
        function popupInfo(url,height,width,tag) {
           var prefs = "scrollbars=yes,menubars=no,toolbars=no,"
           + "resizable=yes,left=0,top=0,height="+height+",width="+width;
           window.open(url, tag, prefs);
        }
        function toggleAll_jb_add(checked) {
           len = document.jb_add_form.elements.length;
           var i;
           for (i = 0; i < len; i++) {
              document.jb_add_form.elements[i].checked = checked;
           }
        }
        function toggleAll_jb_search_pl(checked) {
           len = document.jb_search_pl_form.elements.length;
           var i;
           for (i = 0; i < len; i++) {
              document.jb_search_pl_form.elements[i].checked = checked;
           }
        }
     //-->
     </script>
   }
   proc init { } {
     return 1
   }
}
###############################################################################
# Function   : jukebox::process state machine
# Description: Main function that is called by web server when /jukebox
#              is entered.
#              e.g. /jukebox?jb_search_pl=*abc*
# Author     : Tom Wilkason
# Date       : 12/15/2002
###############################################################################
proc jukebox::process {sock {params ""}} {
   Trace
   #puts "jukebox::process $sock=>$params"
   variable search_form
   variable javascript
   upvar #0 Httpd$sock data
   set title "SnackAmp Jukebox"
   puts $sock [subst $control::head]
   puts $sock [subst {<a href="/jukebox?$params">Refresh</a>|}]
   puts $sock {<a href="/display" target=saplaylist>Display Playlist</a>|}
   #puts $sock {<a href="javascript:popupInfo('/display',600,600,'playlist')">Display Playlist</a>|}
   puts $sock $jukebox::close_window
   ;##
   ;# Need to have search_form optional
   ;#
   #puts $sock [subst $search_form]
   puts $sock $javascript
   set Search_What pl
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      #puts "jukebox::process => '$CmdType' = '$Cmd'"
      switch -- $CmdType {
         jb_search {
            ;# stage 1
            set filter $Cmd
         }
         search_what {
            ;# stage 2
            switch -glob -- [string tolower $Cmd] {
               "all*files" {
                  ;# User specified text to search catalog
                  puts $sock "<h2>The following tracks matched your search criteria '$filter'</h2>"
                  puts $sock "<h2>Please check desired tracks and press <i>Add</i> at the bottom to add them to the playlist</h2>"
                  ;# Pass params for putting in back button on next form
                  puts $sock [jukebox::searchCatalog $filter $::snackAmpSettings(httpNextLength) $params]
               }
               "play*" {
                  ;# User specifed some text to search within the playlist
                  ;# This goes in a pop-up
                  puts $sock "<h2>The following tracks from the Play List matched your search criteria '$filter'</h2>"
                  puts $sock [jukebox::searchPlaylist $filter 0]
               }
               "auto*" {
                  ;# User specifed some text to search within the playlist
                  ;# This goes in a pop-up
                  puts $sock "<h2>The following tracks from the Auto-Playlist matched your search criteria '$filter'</h2>"
                  puts $sock [jukebox::searchPlaylist $filter 1]
               }
            } ;# End Switch
         }
         play {
            ;# Play a specific song
            ;# /control?play=some+name.mp3
            if {[catch {filtering::InsertPlay $Cmd} error] } then {
               puts stderr $result
               puts $sock $result
            }
         }
         delete {
            LB::delete $Cmd
            updateFileDisplay
            ;# Remove from playlist
         }
         question {
            ;# This goes in a pop-up
            ;# javascript pop-up track details (like playinfo)
            if {![file exists $Cmd]} {
               ;# TFW: TF Lookup needs to be wrapped in LB
               set Cmd [LB::getFullNameFromHash $Cmd 1]
            } ;# end if
            ;# TFW: would be nice to make the fonts smaller
            puts $sock [control::trackInfo $Cmd]
         }
         jb_add_start {
            ;# When processing a playlist add, this is the first param
            set playing [LB::active]
            set old_end [LB::end]
            ;# The back reference uses the jb_add_start param (which is the params from the
            ;# request that originated the search) which was pass in here. It is saved
            ;# In a hidden field from the prior form.
            puts $sock [subst {|<a href="/jukebox?$Cmd">Back</a>}]
            puts $sock "<h2>Added the following tracks to the playlist</h2>"
            puts $sock {<table border="1" width="100%" cellpadding="0" cellspacing="0"}
         }
         jb_add {
            ;# When processing a playlist add, these are the middle params, one per item to add
            set track [file root [file tail $Cmd]]
            set folder [file tail [file dirname $Cmd]]
            puts $sock "<tr><td>$folder</td><td>$track</td></tr></p>"
            macros::Append $Cmd
         }
         jb_add_process {
            ;# When processing a playlist add, this is the final param, signals end of list
            puts $sock "</table>"
            puts $sock "<h2>Your requested tracks were added at the end, [expr {$old_end - $playing}] tracks to go</h1>"
            updateFileDisplay
         }
      }
   }
   ;# Insert Display Playlist link
   puts $sock "</body>\n</html>\n"
   HttpD::SocketDone $sock 1
   return 0
}
###############################################################################
# Function   : getAPLFileList
# Description: Return a list of all APL files
# Author     : Tom Wilkason
# Date       : 12/24/2002
###############################################################################
proc jukebox::getAPLFileList {args} {
   #set RF $snackAmpSettings(UseRemovableFolders)
   #set rmtFilter [file join $::RMT *]
   ;# TFW: Add proper support for removable media here
   ;##
   set result [list]
   foreach {File} [catalog::sortedList] {
      if {[filtering::checkIfIncluded [file dirname $File]]} {
         lappend result $File
      }
   }
   return $result
}
###############################################################################
# Function   : question
# Description: Question Button, pop-up info on a song 
#              Handling this two encoding is a real pain in the ass
#              still won't handle a + in file name
# Author     : Tom Wilkason
# Date       : 4/24/2002
###############################################################################
proc jukebox::questionButton {ref height tag} {
   set ref [HttpD::Encode $ref]
   return [subst {<td><a href="/jukebox?question=$ref" target=sainfo><img border="0" src="/remote/images/questionM.gif"></a></td>}]

   ;# Need to doubly encode the reference since there are two steps of decode done
   ;# However, double encode only works on windows/IE
   ;# basically, this javascript shit won't work cross-platform
   ;# so we just pop-up a plain old browser
   set ref [HttpD::Encode [string map {' \\'} $ref]]
   #set ref [HttpD::Encode $ref]
   #set ref [HttpD::Encode [HttpD::Encode $ref]]
   return [subst {<td><a href="javascript:popupInfo('/jukebox?question=$ref',$height,600,'$tag')"><img border="0" src="/remote/images/questionM.gif"></a></td>}]
   #return [subst {<td><a href="javascript:popupInfo('/jukebox?question='+escape('$ref'),$height,600,'$tag')"><img border="0" src="/remote/images/questionM.gif"></a></td>}]
}
###############################################################################
# Function   : searchPlaylist
# Description: Based on a pattern from the web entry box, filter the current
#              playlist and return the matching results
#
# Author     : Tom Wilkason
###############################################################################
proc jukebox::searchPlaylist {filter {aplOnly}} {
   set i 0
   set result {<table border="1" width="100%" cellpadding="0" cellspacing="0">}
   if {$aplOnly} {
      set files [jukebox::getAPLFileList]
   } else {
      set files [LB::getAllNames 1]
   } ;# end if
   set j 0
   foreach {file} $files {
      if {[string match -nocase *$filter* $file]} {
         set bgc $control::row_colors([expr {($j/5)&0x1}])
         #set ref [HttpD::Encode $file]
         #append result "<tr><td><a href=\"/control?play=$ref\">$file</a></td></tr>\n"
         set track [file root [file tail $file]]
         set folder [file tail [file dirname $file]]
         if {$folder==""} {
            set folder "&nbsp;"
         } ;# end if
         set btn [jukebox::questionButton $file 200 "info"]
         ;# The second param just performs another search using the same filter
         append btn [subst {<td><a href="/jukebox?play=[HttpD::Encode $file]&jb_search=$filter&search_what=Playlist"><img border="0" src="/remote/images/playM.gif"></a></td>}]
         append btn [subst {<td><a href="/jukebox?delete=$i&jb_search=$filter&search_what=Playlist"><img border="0" src="/remote/images/deleteM.gif"></a></td>}]
         append result "<tr bgcolor=$bgc><td>$i</td>$btn<td>$folder</td><td>$track</td></tr>\n"
         incr j
      } ;# end if
      incr i
   } ;# end foreach
   append result "</table>"
   #append result "<a href=\"javascript:toggleAll_jb_search_pl(true)\">Select All</a> | <a href=\"javascript:toggleAll_jb_search_pl(false)\">Unselect All"
   return $result
}

###############################################################################
# Function   : jukebox::searchCatalog
# Description: Search all songs (up to N) for results
#              Returns HTML shown all the tracks that matched the filter
#              Info and checkboxes are included
# Author     : Tom Wilkason
# Date       : 12/22/2002
###############################################################################
proc jukebox::searchCatalog {filter max parms} {
   set i 0
   append result "<FORM METHOD=\"POST\" ACTION=\"/jukebox\" NAME=\"jb_add_form\">"
   append result "<input type=hidden name=jb_add_start value=$parms>"
   ;# Initial param will be jb_add_start, followed by each checked item as jb_add, ending with jb_add_process=Submit
   append result {<table border="1" width="100%" cellpadding="0" cellspacing="0">}
   foreach {file} [catalog::sortedList] {
      if {[string match -nocase *$filter* $file]} {
         set bgc $control::row_colors([expr {($i/5)&0x1}])
         set track [file root [file tail $file]]
         set folder [file tail [file dirname $file]]
         if {$folder==""} {
            set folder "&nbsp;"
         } ;# end if
         set btn [jukebox::questionButton $file 200 "info"]
         ;# The second param just performs another search using the same filter
         append btn [subst {<td><a href="/jukebox?play=[HttpD::Encode $file]&jb_search=$filter&search_what=AllFiles"><img border="0" src="/remote/images/playM.gif"></a></td>}]
         append result "<tr bgcolor=$bgc><td><INPUT TYPE=checkbox NAME=\"jb_add\" VALUE=\"$file\"></td>$btn<td>$folder</td><td>$track</td></tr>\n"
         incr i
      } ;# end if
      if {$i>$max} {
         break
      } ;# end if
   } ;# end foreach
   append result "</table>\n"
   if {$i>$max} {
      append result "<p>Results limited to $max matches</p>"
   } ;# end if
   append result "<INPUT TYPE=submit VALUE=Add NAME=jb_add_process>\n"
   append result "</FORM>\n"
   append result "<a href=\"javascript:toggleAll_jb_add(true)\">Select All</a> | <a href=\"javascript:toggleAll_jb_add(false)\">Unselect All<br>"
   return $result
}

set httpToDo {
- Playlists on play_control
- Ability to change the queued print out size
- FF/FR to play control
- Figure a way to push data to browser without it requesting it
  Have knowlege when a socket is open.
- Clicking on a song should bring up a pop-up with song details, use
  play button to play a song (can use javascript like my pop-up)
- Animated gif to denote if a song is playing

Unix bugs
- KDE/Netscape the meta refresh doesn't seem to work
- Pausing a song doesn't get an indication on the browser,
  it only knows if it is stopped or playing.
}



;##
;# Set a few defaults and fire up the server
;#
package require HttpD
set snackAmpSettings(webRoot) $TA(webRoot)
set snackAmpSettings(usedPort) [HttpD::Start $snackAmpSettings(port) $::snackAmpSettings(webRoot) $snackAmpSettings(HttpDefault)]
if {[info exists browser::helpFile]} {
   set snackAmpSettings(internalWebBase) "http://[info hostname]:$snackAmpSettings(usedPort)"
   set browser::helpFile "$snackAmpSettings(internalWebBase)/$browser::helpFile"
   set browser::playControl "$snackAmpSettings(internalWebBase)/$::snackAmpSettings(HttpDefault)"
   set browser::Jukebox "$snackAmpSettings(internalWebBase)/jukebox"
} ;# end if

